# $Id: LSCsegFindSegGenerator.py,v 1.8 2005/09/20 18:49:49 kipp Exp $

#
# Preamble.
#

import logging

from glue import lal
from glue import segments
import onasys.SegGenerator
from onasys import onasysd


#
# A single-instrument segment generator based on queries to an LSCsegFind
# server with optional confirmation queries to an LSCdataFind server.
#

class SegGeneratorConfig(onasys.SegGenerator.SegGeneratorConfig, onasysd.SegFindConfig):
	def defaults(self):
		return onasysd.SegFindConfig.defaults(self) + [
			("datafind", "False"),
			("initial_epoch", "100000000")
		]

	def parse(self):
		onasysd.SegFindConfig.parse(self)
		if self.parser.getboolean(self.section, "datafind"):
			self.datafindconfig = onasysd.DataFindConfig(self.filename)
			self.datafindconfig.parse()
			self.datafind = onasysd.DataFind(self.datafindconfig)
		else:
			self.datafind = None
		self.initial_epoch = lal.LIGOTimeGPS(self.parser.get(self.section, "initial_epoch"))

	def log(self):
		l = onasys.SegGenerator.SegGeneratorConfig.log(self)
		if self.datafind:
			l += self.datafindconfig.log()
		return l

	def validate(self):
		if self.datafind:
			self.datafindconfig.validate()


class SegGenerator(onasys.SegGenerator.SegGenerator):
	def __init__(self, config):
		self.config = config
		self.segfind = onasysd.SegFind(config)
		self.search_window = onasysd.SearchWindow()

	def get_state(self):
		return self.search_window

	def set_state(self, state):
		self.search_window = state

	def generate(self, clock):
		# Determine the window in which to search for segments.
		self.search_window.next(self.config.initial_epoch, clock.maxend(), self.config.min_query_length)

		# Put a useful message in the log
		logging.info(self.segfind.command(self.search_window.window))

		# Do an LSCsegFind query
		try:
			list = self.segfind.list(self.search_window.window)
		except Exception, e:
			raise onasys.SegGenerator.SegGeneratorError, "LSCsegFind query failure: " + str(e)
		logging.info("LSCsegFind response = " + str(list))

		# If needed, compute the intersection of the segments and
		# the response of an LSCdataFind query.
		if self.config.datafind:
			try:
				datafindlist = segments.segmentlist()
				for seg in list:
					logging.info(self.config.datafind.command(seg))
					datafindlist |= self.config.datafind.list(seg)
				logging.info("LSCdataFind response = " + str(datafindlist))
				list &= datafindlist
			except Exception, e:
				raise onasys.SegGenerator.SegGeneratorError, "LSCdataFind query failure: " + str(e)

		return {"available": list}
